/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Resource for RADIUS Authentication and Accounting
 * 
 * Copyright (C) 1999, Mark Spencer
 *
 * Mark Spencer <markster@linux-support.net>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <asterisk/lock.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/config.h>
#include <asterisk/app.h>
#include <asterisk/musiconhold.h>
#include <asterisk/options.h>
#include <asterisk/cli.h>
#include <asterisk/say.h>

#include <asterisk/translate.h>
#include <asterisk/parking.h>
#include <asterisk/callerid.h>

#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <pthread.h>

#include <asterisk/res_radius.h>

static char *tdesc = "RADIUS Authentication and Accounting subsystem";

static char *CDR_desc = "RADIUS CDR Backend";
static char *CDR_name = "radius";

static char *set_server_app = "SetRadiusServer";
static char *set_server_synopsis = "Set channel variable 'radius_server' based on the channels source";
static char *set_server_descrip =
"  Usage: SetRadiusServer\n"
"Sets the channels variable 'radius_server' based on the channel source and the configured radius profiles.\n";

static char *dial_tdesc = "RADIUS assisted Dialing Application";
static char *dial_app = "Radius";
static char *dial_synopsis = "Place a call and connect to the current channel using parameters received from the radius server";
static char *dial_descrip =
"  RADIUS(ROUTER/resource[&ROUTER/resource2...][|timeout][|options][|URL]):\n"
"Requests one or more ROUTER channels and places specified outgoing calls on them.\n"
"The ROUTER channel acts like the LOCAL channel, but places the call into a\n"
"[provider-XXXX] context, where XXXX represents a destination route specified\n"
"by the response from the RADIUS server. As soon as a channel answers, the RADIUS\n"
"app will answer the originating channel (if it needs to be answered) and will\n"
"bridge a call with the channel which first answered. All other calls placed by\n"
"the RADIUS app will be hung up. The RADIUS app will then send an accounting START\n"
"recquest to the radius server. If a timeout is not specified, the RADIUS\n"
"application will wait indefinitely until either one of the called channels\n"
"answers, the user hangs up, or all channels return busy or error. In general,\n"
"the dialler will return 0 if it was unable to place the call, or the timeout\n"
"expired. However, if all channels were busy, and there exists an extension with\n"
"priority n+101 (where n is the priority of the dialler instance), then it will\n"
"be the next executed extension (this allows you to setup different behavior on\n"
"busy from no-answer).\n\n"
"  This application returns -1 if the originating channel hangs up, or if the\n"
"call is bridged and either of the parties in the bridge terminate the call.\n"
"When the call is terminated, the RADIUS CDR backend will send an accounting\n"
"STOP record to the RADIUS server.\n\n"
"The option string may contain zero or more of the following characters:\n"
"      't' -- allow the called user transfer the calling user\n"
"      'T' -- to allow the calling user to transfer the call.\n"
"      'r' -- indicate ringing to the calling party, pass no audio until answered.\n"
"      'm' -- provide hold music to the calling party until answered.\n"
"      'H' -- allow caller to hang up by hitting *.\n"
"      'C' -- reset call detail record for this call.\n"
"      'P[(x)]' -- privacy mode, using 'x' as database if provided.\n"
"      'g' -- goes on in context if the destination channel hangs up\n"
"      'A(x)' -- play an announcement to the called party, using x as file\n"
"  In addition to transferring the call, a call may be parked and then picked\n"
"up by another user.\n\n"
"  The optional URL will be sent to the called party if the channel supports\n"
"it.\n\n"
"The RADIUS server parameters are set in the radius.conf file, and can be viewed\n"
"with the RADIUS SHOW CONFIG command\n";

#define DATE_FORMAT "%Y-%m-%d %T"
#define DATE_FORMAT_CISCO_1 "%T.00%%1d %Z %a %b %d %Y"
#define DATE_FORMAT_CISCO_2 "%T.0%%2d %Z %a %b %d %Y"
#define DATE_FORMAT_CISCO_3 "%T.%%3d %Z %a %b %d %Y"
#define STATUS_FORMAT "%s=%s:%s:%s=%s:%s=%s:%s=:%s=0:%s=%s"

static int RadiusRetries;
static char RadiusDir[256] = "";

static char RadiusAuthServer[50] = "";
static char RadiusAuthSecret[50] = "";
static char RadiusAuthType[50] = "";
static int RadiusAuthPort;

static char RadiusAcctServer[50] = "";
static char RadiusAcctSecret[50] = "";
static char RadiusAcctType[50] = "";
static int RadiusAcctPort;

static float RadiusTimeout = 3;
static int tmpPort;
static char *config = "radius.conf";
static int PrePaid;
static char *TimeLimit;

typedef struct radius_server {
	/* Users who can access various contexts */
	char Name[50];		// Name of the server ( from radius.conf section header )
	char Dictionary[50];		// directory for the radius dictionary
	char Server[50];	// server name or IP address
	char Secret[50];	// secret to use for this server
	char Type[50];		// type of server ( currently blank or 'Cisco' )
	int AuthPort;		// authentication port
	int AcctPort;		// accounting port
	int Retries;		// Number of times to retry a RADIUS send
	float Timeout;		// time in seconds to wait for a RADIUS response
	struct radius_server *next;
} RADIUS_SERVER;

static struct ast_radius_list {
	struct radius_server *servers;
	ast_mutex_t lock;
} radius_list = { NULL, AST_MUTEX_INITIALIZER };


static struct radius_server *find_server(char *name)
{
	int found = 0;
	RADIUS_SERVER *s;
	if (name == NULL) return NULL;
	ast_mutex_lock(&radius_list.lock);
	s = radius_list.servers;
	while(s) {
		if (!strcasecmp(s->Name, name)) {
			found++;
			break;
		}
		s = s->next;
	}
	ast_mutex_unlock(&radius_list.lock);
	if (found == 0) {
		ast_log(LOG_DEBUG, "%s does not have a configuration\n", name);
		return NULL;
	}
	return s;
}

static struct radius_server *build_server(char *name, struct ast_variable *v)
{
	struct radius_server *server;
	struct radius_server *prev;
	int found=0;
	prev = NULL;
	ast_mutex_lock(&radius_list.lock);
	server = radius_list.servers;
	while(server) {
		if (!strcasecmp(server->Name, name)) {	
			break;
		}
		prev = server;
		server = server->next;
	}
	if (server) {
		found++;
		/* Already in the list, remove it and it will be added back (or FREE'd) */
		if (prev) {
			prev->next = server->next;
		} else {
			radius_list.servers = server->next;
		}
		ast_mutex_unlock(&radius_list.lock);
 	} else {
		ast_mutex_unlock(&radius_list.lock);
		server = malloc(sizeof(struct radius_server));
		memset(server, 0, sizeof(struct radius_server));
	}
	if (server) {
		if (!found) {
			strncpy(server->Name, name, sizeof(server->Name)-1);
		}
		while(v) {
			if (!strcasecmp(v->name, "Timeout")) {
				server->Timeout = atof(v->value);
				if (server->Timeout == 0) server->Timeout = 1;
			} else if (!strcasecmp(v->name, "Retries")) {
				server->Retries = atoi(v->value);
				if (server->Retries < 1) server->Retries = 1;
			} else 	if (!strcasecmp(v->name, "Server")) {
				strncpy(server->Server, v->value, sizeof(server->Server)-1);
			} else if (!strcasecmp(v->name, "Dictionary")) {
				strncpy(server->Dictionary, v->value, sizeof(server->Dictionary)-1);
			} else if (!strcasecmp(v->name, "Secret")) {
				strncpy(server->Secret, v->value, sizeof(server->Secret)-1);
			} else if (!strcasecmp(v->name, "Type")) {
				strncpy(server->Type, v->value, sizeof(server->Type)-1);
			} else if (!strcasecmp(v->name, "AuthPort")) {
				if (sscanf(v->value, "%i", &tmpPort) == 1) {
					server->AuthPort = atoi(v->value);
				} else {
					ast_log(LOG_ERROR, "Invalid RADIUS AUTH port number '%s' at line %d of %s\n", v->value, v->lineno, config);
					return NULL;
				}
			} else if (!strcasecmp(v->name, "AcctPort")) {
				if (sscanf(v->value, "%i", &tmpPort) == 1) {
					server->AcctPort = atoi(v->value);
				} else {
					ast_log(LOG_ERROR, "Invalid RADIUS AUTH port number '%s' at line %d of %s\n", v->value, v->lineno, config);
					return NULL;
				}
			} //else if (strcasecmp(v->name,"type"))
			//	ast_log(LOG_WARNING, "Ignoring %s\n", v->name);
			v=v->next;
		}
	}
	if ( server->Timeout == 0 ) server->Timeout = RadiusTimeout;
	if ( server->Retries == 0 ) server->Retries = RadiusRetries;
	if ( strlen(server->Server) == 0) {
		ast_log(LOG_ERROR, "Server profile '%s' lacks a server name\n", server->Name);
		return NULL;
	}
	if ( strlen(server->Dictionary) == 0) {
		ast_log(LOG_ERROR, "Server profile '%s' lacks an attribute dictionary\n", server->Name);
		return NULL;
	}
	if ( strlen(server->Secret) == 0) {
		ast_log(LOG_ERROR, "Server profile '%s' lacks a server secret\n", server->Name);
		return NULL;
	}
	return server;
}

/*
 *	Send a RADIUS Accounting packet and receive the reply
 */
static int send_radius_acct_request(RADIUS_SERVER *server, RADIUS_PACKET *req, RADIUS_PACKET **rep)
{
/* returns	-1 = error
		 0 = not authorized
		 1 = authorized
*/
	int i;
	struct timeval	tv;
	char src[64], dst[64];

//ast_log(LOG_NOTICE,"REQUEST ATTRIBUTE PAIRS:\n");
//vp_printlist(stdout, &req->vps);

	for (i = 0; i < server->Retries + 1; i++) {
		fd_set	rdfdesc;

		ip_ntoa(dst, req->dst_ipaddr);
		ast_log(LOG_DEBUG,"  doing Accounting send to %s port %d (try # %d)\n",dst,req->dst_port,i);
		if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "  doing rad_send to %s port %d (try # %d)\n",dst,req->dst_port,i);
		rad_send(req, NULL, server->Secret);

		/* And wait for reply, timing out as necessary */
		FD_ZERO(&rdfdesc);
		FD_SET(req->sockfd, &rdfdesc);

		tv.tv_sec = (int)server->Timeout;
		tv.tv_usec = 1000000 * (server->Timeout - (int)server->Timeout);

		ast_log(LOG_DEBUG,"  waiting for server response (try # %d)\n",i);
		/* Something's wrong if we don't get exactly one fd. */
		if (select(req->sockfd + 1, &rdfdesc, NULL, NULL, &tv) != 1) {
			if (option_verbose > 2) ast_log(LOG_NOTICE,"  rad_send error detected (try # %d)\n",i);
			continue;
		}

		ast_log(LOG_DEBUG,"  receiving response (try # %d)\n",i);
		*rep = rad_recv(req->sockfd);
		if (*rep != NULL) {
			/*
			 *	If we get a response from a machine
			 *	which we did NOT send a request to,
			 *	then complain.
			 */
			if (((*rep)->src_ipaddr != req->dst_ipaddr) ||
			    ((*rep)->src_port != req->dst_port)) {

				ip_ntoa(src, (*rep)->src_ipaddr);
				ip_ntoa(dst, req->src_ipaddr);
				ast_log(LOG_ERROR, "RADIUS response received from '%s' - wrong RADIUS server (should be '%s')\n",src,dst);
				if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "RADIUS response received from '%s' - wrong RADIUS server (should be '%s')\n",src,dst);
				return -1;
			}
			break;
		} else {	/* NULL: couldn't receive the packet */
			ast_log(LOG_DEBUG,"did not receive RADIUS response\n");
			if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "did not receive RADIUS response\n");
			return -1;
		}
	}

	/* No response or no data read (?) */
	if (i == server->Retries + 1) {
		ast_log(LOG_DEBUG, "No response from RADIUS server\n");
		if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "No response from RADIUS server\n");
		return -1;
	}

	if (rad_decode(*rep, req, server->Secret) != 0) {
		ast_log(LOG_DEBUG, "Unable to decode RADIUS response\n");
		if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "Unable to decode RADIUS response\n");
		return -1;
	}

	/* libradius debug already prints out the value pairs for us */
	if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Received response ID %d, code %d, length = %d\n",(*rep)->id, (*rep)->code, (*rep)->data_len);
	vp_printlist(stdout, (*rep)->vps);
	if((*rep)->code == PW_ACCOUNTING_RESPONSE) {
		ast_log(LOG_DEBUG,"RADIUS Accounting packet accepted\n");
		return 1;
	} else {
		ast_log(LOG_DEBUG,"RADIUS Accounting packet NOT ACCEPTED\n");
		if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "RADIUS Accounting packet NOT ACCEPTED\n");
		return 0;
	}

	return 0;
}

/*
 *	Send a RADIUS Authentication packet and receive the reply
 */
static int send_radius_auth_request(RADIUS_SERVER *server, RADIUS_PACKET *req, RADIUS_PACKET **rep)
{
/* returns	-1 = error
		 0 = not authorized
		 1 = authorized
*/
	int i;
	struct timeval	tv;
	char src[64], dst[64];

//ast_log(LOG_NOTICE,"REQUEST ATTRIBUTE PAIRS:\n");
//vp_printlist(stdout, &req->vps);

	for (i = 0; i < server->Retries + 1; i++) {
		fd_set	rdfdesc;

		ip_ntoa(dst, req->dst_ipaddr);
		ast_log(LOG_DEBUG,"  doing Authentication send to %s port %d (try # %d)\n",dst,req->dst_port,i);
		if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "  doing Authentication send to %s port %d (try # %d)\n",dst,req->dst_port,i);
		rad_send(req, NULL, RadiusAuthSecret);

		/* And wait for reply, timing out as necessary */
		FD_ZERO(&rdfdesc);
		FD_SET(req->sockfd, &rdfdesc);

		tv.tv_sec = (int)server->Timeout;
		tv.tv_usec = 1000000 * (server->Timeout - (int)server->Timeout);

		ast_log(LOG_DEBUG,"  waiting for server response (try # %d)\n",i);
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  waiting for server response (try # %d)\n",i);
		/* Something's wrong if we don't get exactly one fd. */
		if (select(req->sockfd + 1, &rdfdesc, NULL, NULL, &tv) != 1) {
			ast_log(LOG_ERROR,"  send error detected (try # %d)\n",i);
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  send error detected (try # %d)\n",i);
			continue;
		}

		ast_log(LOG_DEBUG,"  receiving response (try # %d)\n",i);
		if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "  receiving response (try # %d)\n",i);
		*rep = rad_recv(req->sockfd);
		if (*rep != NULL) {
			/*
			 *	If we get a response from a machine
			 *	which we did NOT send a request to,
			 *	then complain.
			 */
			if (((*rep)->src_ipaddr != req->dst_ipaddr) ||
			    ((*rep)->src_port != req->dst_port)) {

				ip_ntoa(src, (*rep)->src_ipaddr);
				ip_ntoa(dst, req->src_ipaddr);
				ast_log(LOG_ERROR,"RADIUS response received from '%s' - wrong RADIUS server (should be '%s')\n",src,dst);
				if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "RADIUS response received from '%s' - wrong RADIUS server (should be '%s')\n",src,dst);
				return -1;
			}
			break;
		} else {	/* NULL: couldn't receive the packet */
			ast_log(LOG_ERROR, "did not receive RADIUS response\n");
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "did not receive RADIUS response\n");
			return -1;
		}
	}

	/* No response or no data read (?) */
	if (i == server->Retries + 1) {
		ast_log(LOG_ERROR, "No response from RADIUS server\n");
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "No response from RADIUS server\n");
		return -1;
	}

	if (rad_decode(*rep, req, server->Secret) != 0) {
		ast_log(LOG_ERROR, "Unable to decode RADIUS response\n");
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Unable to decode RADIUS response\n");
		return -1;
	}

	/* libradius debug already prints out the value pairs for us */
	ast_log(LOG_DEBUG,"Received response ID %d, code %d, length = %d\n",(*rep)->id, (*rep)->code, (*rep)->data_len);
	if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "Received response ID %d, code %d, length = %d\n",(*rep)->id, (*rep)->code, (*rep)->data_len);
//	vp_printlist(stdout, (*rep)->vps);
	if((*rep)->code == PW_AUTHENTICATION_ACK) {
		ast_log(LOG_DEBUG,"RADIUS Authentication passed\n");
		if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "RADIUS Authentication passed\n");
		return 1;
	} else {
		ast_log(LOG_ERROR,"RADIUS Authentication failed\n");
		if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "RADIUS Authentication failed\n");
		return 0;
	}
	return 0;
}

/* We define a customer "local user" structure because we
   use it not only for keeping track of what is in use but
   also for keeping track of who we're dialing. */

struct localuser {
	struct ast_channel *chan;
	int stillgoing;
	int allowredirect_in;
	int allowredirect_out;
	int ringbackonly;
	int musiconhold;
	int allowdisconnect;
	struct localuser *next;
};

LOCAL_USER_DECL;

static void hanguptree(struct localuser *outgoing, struct ast_channel *exception)
{
	/* Hang up a tree of stuff */
	struct localuser *oo;
	while(outgoing) {
		/* Hangup any existing lines we have open */
		if (outgoing->chan && (outgoing->chan != exception))
			ast_hangup(outgoing->chan);
		oo = outgoing;
		outgoing=outgoing->next;
		free(oo);
	}
}

#define MAX 256

static struct ast_channel *wait_for_answer(struct ast_channel *in, struct localuser *outgoing, int *to, int *allowredir_in, int *allowredir_out, int *allowdisconnect)
{
	struct localuser *o;
	int found;
	int numlines;
	int sentringing = 0;
	int numbusies = 0;
	int orig = *to;
	struct ast_frame *f;
	struct ast_channel *peer = NULL;
	struct ast_channel *watchers[MAX];
	int pos;
	int single;
	int moh=0;
	int ringind=0;
	struct ast_channel *winner;
	
	single = (outgoing && !outgoing->next && !outgoing->musiconhold && !outgoing->ringbackonly);
	
	if (single) {
		/* Turn off hold music, etc */
		ast_indicate(in, -1);
		/* If we are calling a single channel, make them compatible for in-band tone purpose */
		ast_channel_make_compatible(outgoing->chan, in);
	}
	
	if (outgoing) {
		moh = outgoing->musiconhold;
		ringind = outgoing->ringbackonly;
		if (outgoing->musiconhold) {
			ast_moh_start(in, NULL);
		} else if (outgoing->ringbackonly) {
			ast_indicate(in, AST_CONTROL_RINGING);
		}
	}
	
	while(*to && !peer) {
		o = outgoing;
		found = -1;
		pos = 1;
		numlines = 0;
		watchers[0] = in;
		while(o) {
			/* Keep track of important channels */
			if (o->stillgoing && o->chan) {
				watchers[pos++] = o->chan;
				found = 1;
			}
			o = o->next;
			numlines++;
		}
		if (found < 0) {
			if (numlines == numbusies) {
				if (option_verbose > 2)
					ast_verbose( VERBOSE_PREFIX_2 "Everyone is busy at this time\n");
				/* See if there is a special busy message */
				if (ast_exists_extension(in, in->context, in->exten, in->priority + 101, in->callerid)) 
					in->priority+=100;
			} else {
				if (option_verbose > 2)
					ast_verbose( VERBOSE_PREFIX_2 "No one is available to answer at this time\n");
			}
			*to = 0;
			/* if no one available we'd better stop MOH/ringing to */
			if (moh) {
				ast_moh_stop(in);
			} else if (ringind) {
				ast_indicate(in, -1);
			}
			return NULL;
		}
		winner = ast_waitfor_n(watchers, pos, to);
		o = outgoing;
		while(o) {
			if (o->stillgoing && o->chan && (o->chan->_state == AST_STATE_UP)) {
				if (!peer) {
					if (option_verbose > 2)
						ast_verbose( VERBOSE_PREFIX_3 "%s answered %s\n", o->chan->name, in->name);
					peer = o->chan;
					*allowredir_in = o->allowredirect_in;
					*allowredir_out = o->allowredirect_out;
					*allowdisconnect = o->allowdisconnect;
				}
			} else if (o->chan && (o->chan == winner)) {
				if (strlen(o->chan->call_forward)) {
					char tmpchan[256];
					/* Before processing channel, go ahead and check for forwarding */
					if (option_verbose > 2)
						ast_verbose(VERBOSE_PREFIX_3 "Now forwarding %s to '%s@%s' (thanks to %s)\n", in->name, o->chan->call_forward, o->chan->context, o->chan->name);
					/* Setup parameters */
					snprintf(tmpchan, sizeof(tmpchan),"%s@%s", o->chan->call_forward, o->chan->context);
					ast_hangup(o->chan);
					o->chan = ast_request("Local", in->nativeformats, tmpchan);
					if (!o->chan) {
						ast_log(LOG_NOTICE, "Unable to create local channel for call forward to '%s'\n", tmpchan);
						o->stillgoing = 0;
						numbusies++;
					} else {
						if (in->callerid) {
							if (o->chan->callerid)
								free(o->chan->callerid);
							o->chan->callerid = malloc(strlen(in->callerid) + 1);
							strncpy(o->chan->callerid, in->callerid, strlen(in->callerid) + 1);
						}
						if (in->ani) {
							if (o->chan->ani)
								free(o->chan->ani);
							o->chan->ani = malloc(strlen(in->ani) + 1);
							strncpy(o->chan->ani, in->ani, strlen(in->ani) + 1);
						}
						if (ast_call(o->chan, tmpchan, 0)) {
							ast_log(LOG_NOTICE, "Failed to dial on local channel for call forward to '%s'\n", tmpchan);
							o->stillgoing = 0;
							ast_hangup(o->chan);
							o->chan = NULL;
							numbusies++;
						}
					}
					continue;
				}
				f = ast_read(winner);
				if (f) {
					if (f->frametype == AST_FRAME_CONTROL) {
						switch(f->subclass) {
					    case AST_CONTROL_ANSWER:
							/* This is our guy if someone answered. */
							if (!peer) {
								if (option_verbose > 2)
									ast_verbose( VERBOSE_PREFIX_3 "%s answered %s\n", o->chan->name, in->name);
								peer = o->chan;
								*allowredir_in = o->allowredirect_in;
								*allowredir_out = o->allowredirect_out;
								*allowdisconnect = o->allowdisconnect;
							}
							break;
						case AST_CONTROL_BUSY:
							if (option_verbose > 2)
								ast_verbose( VERBOSE_PREFIX_3 "%s is busy\n", o->chan->name);
							in->hangupcause = o->chan->hangupcause;
							ast_hangup(o->chan);
							o->chan = NULL;
							o->stillgoing = 0;
							if (in->cdr)
								ast_cdr_busy(in->cdr);
							numbusies++;
							break;
						case AST_CONTROL_CONGESTION:
							if (option_verbose > 2)
								ast_verbose( VERBOSE_PREFIX_3 "%s is circuit-busy\n", o->chan->name);
							in->hangupcause = o->chan->hangupcause;
							ast_hangup(o->chan);
							o->chan = NULL;
							o->stillgoing = 0;
							if (in->cdr)
								ast_cdr_busy(in->cdr);
							numbusies++;
							break;
						case AST_CONTROL_RINGING:
							if (option_verbose > 2)
								ast_verbose( VERBOSE_PREFIX_3 "%s is ringing\n", o->chan->name);
							if (!sentringing && !moh) {
								ast_indicate(in, AST_CONTROL_RINGING);
								sentringing++;
								ringind++;
							}
							break;
						case AST_CONTROL_PROGRESS:
							if (option_verbose > 2)
								ast_verbose ( VERBOSE_PREFIX_3 "%s is making progress passing it to %s\n", o->chan->name,in->name);
							ast_indicate(in, AST_CONTROL_PROGRESS);
							break;
						case AST_CONTROL_OFFHOOK:
							/* Ignore going off hook */
							break;
						case -1:
							if (option_verbose > 2)
								ast_verbose( VERBOSE_PREFIX_3 "%s stopped sounds\n", o->chan->name);
							ast_indicate(in, -1);
							break;
						default:
							ast_log(LOG_DEBUG, "Dunno what to do with control type %d\n", f->subclass);
						}
					} else if (single && (f->frametype == AST_FRAME_VOICE) && 
								!(outgoing->ringbackonly || outgoing->musiconhold)) {
						if (ast_write(in, f)) 
							ast_log(LOG_WARNING, "Unable to forward frame\n");
					} else if (single && (f->frametype == AST_FRAME_IMAGE) && 
								!(outgoing->ringbackonly || outgoing->musiconhold)) {
						if (ast_write(in, f))
							ast_log(LOG_WARNING, "Unable to forward image\n");
					}
					ast_frfree(f);
				} else {
					in->hangupcause = o->chan->hangupcause;
					ast_hangup(o->chan);
					o->chan = NULL;
					o->stillgoing = 0;
				}
			}
			o = o->next;
		}
		if (winner == in) {
			f = ast_read(in);
#if 0
			if (f && (f->frametype != AST_FRAME_VOICE))
					printf("Frame type: %d, %d\n", f->frametype, f->subclass);
			else if (!f || (f->frametype != AST_FRAME_VOICE))
				printf("Hangup received on %s\n", in->name);
#endif
			if (!f || ((f->frametype == AST_FRAME_CONTROL) && (f->subclass == AST_CONTROL_HANGUP))) {
				/* Got hung up */
				*to=-1;
				return NULL;
			}
			if (f && (f->frametype == AST_FRAME_DTMF) && *allowdisconnect &&
				(f->subclass == '*')) {
			    if (option_verbose > 3)
				ast_verbose(VERBOSE_PREFIX_3 "User hit %c to disconnect call.\n", f->subclass);
				*to=0;
				return NULL;
			}
			if (single && ((f->frametype == AST_FRAME_VOICE) || (f->frametype == AST_FRAME_DTMF)))  {
				if (ast_write(outgoing->chan, f))
					ast_log(LOG_WARNING, "Unable to forward voice\n");
				ast_frfree(f);
			}
		}
		if (!*to && (option_verbose > 2))
			ast_verbose( VERBOSE_PREFIX_3 "Nobody picked up in %d ms\n", orig);
	}
	if (moh) {
		ast_moh_stop(in);
	} else if (ringind) {
		ast_indicate(in, -1);
	}

	return peer;
	
}

static int set_radius_server(struct ast_channel *chan, void *data)
{
	int res;
	struct radius_server *use_server;
	char server_name[256];
	
	if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Checking for valid RADIUS Profile...");
	use_server = find_server(pbx_builtin_getvar_helper(chan,"SIPDOMAIN"));
	if (!use_server) {
		use_server = find_server(pbx_builtin_getvar_helper(chan,"radius_profile"));
		if (!use_server) {
			use_server = find_server(chan->context);
			if (!use_server) {
				ast_log(LOG_ERROR,"No matching RADIUS server configured for this channel\n");
				if (option_verbose > 2) ast_verbose( "No matching RADIUS server configured for this channel!\n");
				return -1;
			}
		}
	}
	pbx_builtin_setvar_helper(chan,"radius_profile",use_server->Name);
	if (option_verbose > 2) ast_verbose( "Set 'radius_profile' to '%s'\n\n",use_server->Name);
	return 0;
}

static int RDial_exec(struct ast_channel *chan, void *data)
{
	RADIUS_PACKET *req;
	RADIUS_PACKET *rep = NULL;
	LRAD_TOKEN last_token;
	VALUE_PAIR *list;
	struct radius_server *use_server;
	int res=-1;
	struct localuser *u;
	char info[256], *peers, *timeout, *tech, *number, *rest, *cur;
	char  privdb[256] = "", *s;
	char  announcemsg[256] = "", *ann;
	struct localuser *outgoing=NULL, *tmp;
	struct ast_channel *peer;
	int to;
	int allowredir_in=0;
	int allowredir_out=0;
	int allowdisconnect=0;
	int privacy=0;
	int announce=0;
	int resetcdr=0;
	char numsubst[AST_MAX_EXTENSION];
	char restofit[AST_MAX_EXTENSION];
	char *transfer = NULL;
	char *newnum;
	char callerid[256], *l, *n;
	char *url=NULL; /* JDG */
	struct ast_var_t *current;
	struct varshead *headp;
	int go_on=0;
	int id;
	char radiusQuery[1024]="";
	char *p;
	
	struct tm tm;
	time_t t;
	char time_start[80], time_tmp[80];

	int radius_allocated = 0;
	char radiusQuery_start[1024]="";
	char radius_info[1024], route_provider[10], area_idx[10], route_wholesale[10];
	char route_cost[10], rate_plan[10], billing_unit[10], billing_increment[10];
	char call_status[1024];
	char cdr_userfield[1024];
	char *ptr, *ptr1;
	VALUE_PAIR *list_start;

ast_log(LOG_ERROR, "Starting RDial for call routing on channel '%s'\n", chan->name);
	if (!data) {
		ast_log(LOG_ERROR, "RDial requires an argument (ROUTER/number1&ROUTER/number2...|optional timeout)\n");
		return -1;
	}
	
	LOCAL_USER_ADD(u);
	
	strncpy(info, (char *)data, sizeof(info) - 1);
	peers = info;
	if (peers) {
		timeout = strchr(info, '|');
		if (timeout) {
			*timeout = '\0';
			timeout++;
			transfer = strchr(timeout, '|');
			if (transfer) {
				*transfer = '\0';
				transfer++;
				/* JDG */
				url = strchr(transfer, '|');
				if (url) {
					*url = '\0';
					url++;
					ast_log(LOG_DEBUG, "RDIAL WITH URL=%s_\n", url);
				} else 
					ast_log(LOG_DEBUG, "SIMPLE RDIAL (NO URL)\n");
				/* /JDG */
			}
		}
	} else
		timeout = NULL;
	if (!peers || !strlen(peers)) {
		ast_log(LOG_WARNING, "RDial argument takes format (ROUTER/number1&ROUTER/number2...|optional timeout)\n");
		res = -1;
		goto out;
	}
	

	if (transfer) {
		/* XXX ANNOUNCE SUPPORT */
		if ((ann = strstr(transfer, "A("))) {
			announce = 1;
			strncpy(announcemsg, ann + 2, sizeof(announcemsg) - 1);
			/* Overwrite with X's what was the announce info */
			while(*ann && (*ann != ')')) 
				*(ann++) = 'X';
			if (*ann)
				*ann = 'X';
			/* Now find the end of the privdb */
			ann = strchr(announcemsg, ')');
			if (ann)
				*ann = '\0';
			else {
				ast_log(LOG_WARNING, "Transfer with Announce spec lacking trailing ')'\n");
				announce = 0;
			}
		}
		/* Extract privacy info from transfer */
		if ((s = strstr(transfer, "P("))) {
			privacy = 1;
			strncpy(privdb, s + 2, sizeof(privdb) - 1);
			/* Overwrite with X's what was the privacy info */
			while(*s && (*s != ')')) 
				*(s++) = 'X';
			if (*s)
				*s = 'X';
			/* Now find the end of the privdb */
			s = strchr(privdb, ')');
			if (s)
				*s = '\0';
			else {
				ast_log(LOG_WARNING, "Transfer with privacy lacking trailing ')'\n");
				privacy = 0;
			}
		} else if (strchr(transfer, 'P')) {
			/* No specified privdb */
			privacy = 1;
		} else if (strchr(transfer, 'C')) {
			resetcdr = 1;
		}
	}
	if (resetcdr && chan->cdr)
		ast_cdr_reset(chan->cdr, 0);
	if (!strlen(privdb) && privacy) {
		/* If privdb is not specified and we are using privacy, copy from extension */
		strncpy(privdb, chan->exten, sizeof(privdb) - 1);
	}
//	if (privacy) {
		if (chan->callerid)
			strncpy(callerid, chan->callerid, sizeof(callerid));
		else
			strcpy(callerid, "");
		ast_callerid_parse(callerid, &n, &l);
		if (l) {
			ast_shrink_phone_number(l);
		} else
			l = "";
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Privacy DB is '%s', privacy is %d, clid is '%s'\n", privdb, privacy, l);
//	}
	cur = peers;
	do {
		/* Remember where to start next time */
		rest = strchr(cur, '&');
		if (rest) {
			*rest = 0;
			rest++;
		}
		/* Get a technology/[device:]number pair */
		tech = cur;
		number = strchr(tech, '/');
		if (!number) {
			ast_log(LOG_ERROR, "RDial argument takes format (ROUTER/[device:]number1&ROUTER/[device:]number2...|optional timeout)\n");
			res = -1;
			goto out;
		}
		*number = '\0';
		number++;
		if (strcasecmp(tech,"ROUTER")) {
			ast_log(LOG_ERROR, "RDial argument takes format (ROUTER/[device:]number1&ROUTER/[device:]number2...|optional timeout)\n");
			res = -1;
			goto out;
		}
		tmp = malloc(sizeof(struct localuser));
		if (!tmp) {
			ast_log(LOG_ERROR, "Out of memory\n");
			res = -1;
			goto out;
		}
		memset(tmp, 0, sizeof(struct localuser));
		if (transfer) {
			if (strchr(transfer, 't'))
				tmp->allowredirect_in = 1;
                        else    tmp->allowredirect_in = 0;
			if (strchr(transfer, 'T'))
				tmp->allowredirect_out = 1;
                        else    tmp->allowredirect_out = 0;
			if (strchr(transfer, 'r'))
				tmp->ringbackonly = 1;
                        else    tmp->ringbackonly = 0;
			if (strchr(transfer, 'm'))
				tmp->musiconhold = 1;
                        else    tmp->musiconhold = 0;
			if (strchr(transfer, 'H'))
				allowdisconnect = tmp->allowdisconnect = 1;
                        else    allowdisconnect = tmp->allowdisconnect = 0;
			if(strchr(transfer, 'g'))
				go_on=1;
		}
		strncpy(numsubst, number, sizeof(numsubst)-1);
		/* If we're dialing by extension, look at the extension to know what to dial */
		if ((newnum = strstr(numsubst, "BYEXTENSION"))) {
			strncpy(restofit, newnum + strlen("BYEXTENSION"), sizeof(restofit)-1);
			snprintf(newnum, sizeof(numsubst) - (newnum - numsubst), "%s%s", chan->exten,restofit);
			if (option_debug)
				ast_log(LOG_DEBUG, "Dialing by extension %s\n", numsubst);
		}

//---------------------------------------------------------------------------------------------------------------

		/* build the radius authentication request and send it to the radius server */
		if (chan->cdr) 
			if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_2 "UNIQUEID for this call: '%s'\n", chan->cdr->uniqueid);
		if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "Starting RADIUS Authentication...\n");

		// find the radius server to use
		// check against channel variables 'radius_profile' or 'SIPDOMAIN' for matches to the configured servers
		// then checks chan->context
ast_verbose( VERBOSE_PREFIX_3 "\nChecking for valid RADIUS Profile\n");
		use_server = find_server(pbx_builtin_getvar_helper(chan,"SIPDOMAIN"));
		if (!use_server) {
			use_server = find_server(pbx_builtin_getvar_helper(chan,"radius_profile"));
			if (!use_server) {
				use_server = find_server(chan->context);
				if (!use_server) {
					ast_log(LOG_ERROR,"No matching RADIUS server configured for this channel\n");
					ast_verbose( VERBOSE_PREFIX_1 "No matching RADIUS server configured for this channel\n");
					res = -1;
					goto out;
				}
			}
		}
ast_verbose( VERBOSE_PREFIX_3 "\nUsing RADIUS profile [%s]\n\n",use_server->Name);
pbx_builtin_setvar_helper(chan,"radius_profile",use_server->Name);
		// initialize the RADIUS attribute dictionary
		if (dict_init(use_server->Dictionary, RADIUS_DICTIONARY) < 0) {
			ast_log(LOG_ERROR, "Unable to initialize RADIUS attribute dictionary\n");
			res = -1;
			goto out;
		}
		if ((req = rad_alloc(1)) == NULL) {
			ast_log(LOG_ERROR, "Unable to allocate memory for RADIUS AUTH request for channel of type '%s'\n", tech);
			res = -1;
			goto out;
		} else {
			radius_allocated = 1;
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3  "  RADIUS AUTH Packet allocated...\n");
		}
		id = ((int)getpid() & 0xff);
		req->id = id;
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3  "  RADIUS AUTH request id set to '%d'\n",req->id);
		req->dst_port = use_server->AuthPort;
		req->code = PW_AUTHENTICATION_REQUEST;
		req->dst_ipaddr = ip_getaddr(use_server->Server);
		if (req->dst_ipaddr == INADDR_NONE) {
			ast_log(LOG_WARNING, "Unable to resolve address for RADIUS AUTH server '%s'\n", use_server->Server);
			res = -1;
			goto out;
		} else {
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  RADIUS AUTH server resolved\n");
		}

		if (strlen(chan->accountcode)) {
			// this call needs to be authenticated against an account code
ast_verbose( VERBOSE_PREFIX_3 "Building Auth Query to authenticate against accountcode\n");
			snprintf(radiusQuery, sizeof(radiusQuery), 
			"user-name=%s, 
			calling-station-id=%s, 
			called-station-id=%s\n", 
			l, 
			chan->accountcode, 
			numsubst);
		} else {
ast_verbose( VERBOSE_PREFIX_3 "Building Auth Query to authenticate against ANI\n");
			snprintf(radiusQuery, sizeof(radiusQuery), 
			"user-name=%s, 
			calling-station-id=%s, 
			called-station-id=%s\n", 
			l, 
			l, 
			numsubst);
		}
		
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "\nRADIUS AUTH QUERY:\n%s", radiusQuery);
		list = NULL;
		p = radiusQuery;
		do {
			pairadd(&list, pairread(&p,&last_token));
		} while ( last_token != T_EOL );
		req->vps = list;
		if ((req->sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			ast_log(LOG_ERROR, "Unable to allocate socket for RADIUS AUTH request\n");
			res = -1;
			goto out;
		} else {
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  Socket allocated for RADIUS AUTH request\n");
		}
		librad_md5_calc(req->vector, req->vector, sizeof(req->vector));
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  request vector set for RADIUS AUTH request\n");
		res = send_radius_auth_request(use_server,req, &rep);
		if (!res) {
			// radius authentication failed
			ast_log(LOG_ERROR,"Unable to authenticate call with RADIUS AUTH server!\n");
			if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "Unable to authenticate call with RADIUS AUTH server!\n");
			res = -1;
			goto out;
		} else {
			if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "RADIUS AUTH server Authenticated call\n");
		}
		if (rep) {
			// parse the radius response and convert to Asterisk Variables
			if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "parsing RADIUS AUTH response\n");

			char *p;
			char buf[1024];
			VALUE_PAIR *list;

			list = rep->vps;
			for (; list; list = list->next) {
				// list->name holds the attribute name
				// list->strvalue holds the value (for cisco NAS also holds attribute name)
				if (!strcasecmp(use_server->Type, "Cisco")) {
					//we need to parse the attribute name out of the attribute value for Cisco GW compatability
					p = strstr(list->strvalue, "=");
					strncpy(buf, p + 1, sizeof(buf) - 1);
				} else {
					strncpy(buf, list->strvalue, sizeof(buf) - 1);
				}
				pbx_builtin_setvar_helper(chan, list->name, buf);
				if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "parsed '%s' attribute - value='%s'\n",list->name,buf);
			}

			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "RADIUS AUTH response parsed\n");
			rad_free(&rep);
		} else {
			ast_log(LOG_ERROR,"Unable to authenticate call with RADIUS AUTH server!\n");
			res = -1;
			goto out;
		}
		// free the RADIUS request and close the socket
		pairfree(&req->vps);
		close(req->sockfd);
		rad_free(&req);
		req = NULL;
		radius_allocated = 0;
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Radius AUTH packet de-allocated\n");

		// dump channel variables
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "BILLING MODEL = %s\n",pbx_builtin_getvar_helper(chan,"h323-billing-model"));
		headp=&chan->varshead;
		AST_LIST_TRAVERSE(headp,current,entries) {
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Channel variable '%s' = '%s'\n",ast_var_name(current),ast_var_value(current));
		}

		// check for key radius variables - h323-billing-model, h323-credit-time, h323-ivr-in
		if ( !pbx_builtin_getvar_helper(chan, "h323-billing-model") || !pbx_builtin_getvar_helper(chan, "h323-credit-time") || !pbx_builtin_getvar_helper(chan, "h323-billing-model")) {
			ast_log(LOG_ERROR,"RADIUS server did not return the required parameters to complete the call\n");
			res = -1;
			goto out;
		}
		PrePaid = 0;
		if (strcasecmp(pbx_builtin_getvar_helper(chan, "h323-billing-model"),"Prepaid")==0) {
			PrePaid = 1;
		}
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Call model set to Prepaid=%d\n",PrePaid);

		TimeLimit = pbx_builtin_getvar_helper(chan, "h323-credit-time");
		if ( PrePaid == 1) {
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Call TimeLimit set to %s\n",TimeLimit);
		} else {
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Call TimeLimit set to %s\n","Unlimited");
		}

/*	parse response from RADIUS server ( h323-ivr-in )
		h323-ivr-in - RouteProvider:area_idx:route_wholesale=route_cost:rate_plan:billing_unit=billing_increment
		where	RouteProvider		= the route prefix to use (Least Cost Routing)
			area_idx		= the route index to the routes database entry
			route_wholesale		= the cost per minute for this route
			route_cost		= the retail cost per minute for this route
			rate_plan		= the rate_plan for the customer
			billing_unit		= minimum call time
			billing_increment	= call time increment
*/

		//parse route_provider
		ptr = strncpy(radius_info,pbx_builtin_getvar_helper(chan, "h323-ivr-in"),sizeof(radius_info)-1);
		ptr1 = strchr(ptr, ':');
		if (ptr1)
			*ptr1 = '\0';
		strncpy(route_provider, ptr, sizeof(route_provider) - 1);
		ptr = ptr1 + 1;
		//parse area_idx
		ptr1 = strchr(ptr, ':');
		if (ptr1) 
			*ptr1 = '\0';
		strncpy(area_idx, ptr, sizeof(area_idx) - 1);
		ptr = ptr1 + 1;
		//parse route_wholesale
		ptr1 = strchr(ptr, '=');
		if (ptr1) 
			*ptr1 = '\0';
		strncpy(route_wholesale, ptr, sizeof(route_wholesale) - 1);
		ptr = ptr1 + 1;
		//parse route_cost
		ptr1 = strchr(ptr, ':');
		if (ptr1) 
			*ptr1 = '\0';
		strncpy(route_cost, ptr, sizeof(route_cost) - 1);
		ptr = ptr1 + 1;
		//parse rate_plan
		ptr1 = strchr(ptr, ':');
		if (ptr1) 
			*ptr1 = '\0';
		strncpy(rate_plan, ptr, sizeof(rate_plan) - 1);
		ptr = ptr1 + 1;
		//parse billing_unit & billing_increment
		ptr1 = strchr(ptr, '=');
		if (ptr1)
			*ptr1 = '\0';
		strncpy(billing_unit, ptr, sizeof(billing_unit) - 1);
		strncpy(billing_increment, ptr1+1, sizeof(billing_increment) - 1);
		if ( strlen(billing_unit) == 0 ) {
			strncpy(billing_unit, "30", sizeof(billing_unit) - 1);
		}
		if ( strlen(billing_increment) == 0 ) {
			strncpy(billing_increment, "6", sizeof(billing_increment) - 1);
		}

//ast_log(LOG_NOTICE,"Parsed RADIUS h323-ivr-in\n  Route_Provider:\t'%s'\n  area_idx:\t'%s'\n  route_wholesale:\t'%s'\n  route_cost:\t'%s'\n  rate_plan:\t'%s'\n  billing_unit:\t'%s'\n  billing_increment:\t'%s'\n",route_provider,area_idx,route_wholesale,route_cost,rate_plan,billing_unit,billing_increment);
//ast_log(LOG_NOTICE, "Original h323-ivr-in: '%s'\n", pbx_builtin_getvar_helper(chan, "h323-ivr-in"));

//---------------------------------------------------------------------------------------------------------------

		/* build the radius accounting START request and send it to the radius server */
		// because Asterisk only sends CDRs to the CDR backend at the completion of
		// a call, and RADIUS servers require a START and a STOP record, we need to
		// send both in sequence.
		if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "Starting RADIUS START Accounting...\n");
	
		if ((req = rad_alloc(1)) == NULL) {
			ast_log(LOG_ERROR, "Unable to allocate memory for RADIUS Accounting START request\n");
			return -1;
		} else {
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  RADIUS Accounting START Packet allocated...\n");
		}
		id = ((int)getpid() & 0xff);
		req->id = id;
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  RADIUS Accounting START request id set to '%d'\n",req->id);
		req->dst_port = use_server->AcctPort;
		req->code = PW_ACCOUNTING_REQUEST;
		req->dst_ipaddr = ip_getaddr(use_server->Server);
		if (req->dst_ipaddr == INADDR_NONE) {
			ast_log(LOG_ERROR, "Unable to resolve address for RADIUS Accounting server '%s'\n", use_server->Server);
			return -1;
		} else {
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  RADIUS Accounting server resolved\n");
		}

		// set the call status for the call (to place in h323-return-code
		snprintf(call_status,sizeof(call_status),STATUS_FORMAT,"h323-return-code=VoIP","CALL SETUP STARTED",route_provider,"1502",area_idx,route_wholesale,route_cost,rate_plan,chan->cdr->dst,billing_unit,billing_increment);

		// convert the start time to string
		if (!strcasecmp(use_server->Type, "Cisco")) {
			t = chan->cdr->start.tv_sec;
			localtime_r(&t,&tm);
			if ((chan->cdr->start.tv_usec/1000) < 10) {
				strftime(time_tmp, sizeof(time_tmp), DATE_FORMAT_CISCO_1, &tm);
			} else if ((chan->cdr->start.tv_usec/1000) < 100) {
				strftime(time_tmp, sizeof(time_tmp), DATE_FORMAT_CISCO_2, &tm);
			} else {
				strftime(time_tmp, sizeof(time_tmp), DATE_FORMAT_CISCO_3, &tm);
			}
			snprintf(time_start, sizeof(time_start),time_tmp,chan->cdr->start.tv_usec/1000);
		} else {
			t = chan->cdr->start.tv_sec;
			localtime_r(&t,&tm);
			strftime(time_start, sizeof(time_start), DATE_FORMAT, &tm);
		}

		// build the Radius Start Accounting Query
		if (strlen(chan->cdr->accountcode)) {
			snprintf(radiusQuery_start, sizeof(radiusQuery_start), 
				"Acct-Session-Id=\"%s\", 
				h323-setup-time=\"%s\", 
				user-name=\"%s\", 
				calling-station-id=\"%s\", 
				called-station-id=\"%s\", 
				Acct-Status-Type=\"%s\"\n", 
				chan->cdr->uniqueid, 
				time_start, 
				chan->cdr->accountcode, 
				chan->cdr->src, 
				chan->cdr->dst, 
				"Start");
		} else {
			snprintf(radiusQuery_start, sizeof(radiusQuery_start), 
				"Acct-Session-Id=\"%s\", 
				h323-setup-time=\"%s\", 
				user-name=\"%s\", 
				calling-station-id=\"%s\", 
				called-station-id=\"%s\", 
				Acct-Status-Type=\"%s\"\n", 
				chan->cdr->uniqueid, 
				time_start, 
				chan->cdr->src, 
				chan->cdr->src, 
				chan->cdr->dst, 
				"Start");
		}

// do the START query
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "RADIUS Start QUERY:\n%s", radiusQuery_start);
		list_start = NULL;
		p = radiusQuery_start;
		userparse(p,&list_start);
//	do {
//		pairadd(&list_start, pairread(&p,&last_token));
//	} while ( last_token != T_EOL );
		req->vps = list_start;
		if (option_verbose > 2) {
			ast_verbose( VERBOSE_PREFIX_3 " radius value list is now:\n");
			vp_printlist(stdout, list_start);
			ast_verbose( VERBOSE_PREFIX_3 " radius value pairs are now:\n");
			vp_printlist(stdout, req->vps);
		}
		if ((req->sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			ast_log(LOG_ERROR, "Unable to allocate socket for RADIUS Accounting START request\n");
			return -1;
		} else {
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  socket allocated for RADIUS Accounting START request\n");
		}
		librad_md5_calc(req->vector, req->vector, sizeof(req->vector));
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  request vector set for RADIUS START request\n");
		res = send_radius_acct_request(use_server, req, &rep);
		if (!res) {
			ast_log(LOG_ERROR,"Unable to perform RADIUS Accounting START for call!\n");
			return -1;
		}
		pairfree(&req->vps);
		close(req->sockfd);
		rad_free(&req);
		req = NULL;

//---------------------------------------------------------------------------------------------------------------
		
		/* Request the peer */
		char new_context[256];
		snprintf(new_context,sizeof(new_context),"%s%s","provider-",route_provider);

		char tmpchan[256];
		if (option_verbose > 2)
			ast_verbose(VERBOSE_PREFIX_3 "Forwarding call to '%s@%s'\n", numsubst, new_context);
		snprintf(tmpchan, sizeof(tmpchan),"%s@%s", numsubst, new_context);

snprintf(cdr_userfield, sizeof(cdr_userfield),"AppRDial|%s|%s|%s", chan->uniqueid,use_server->Name,pbx_builtin_getvar_helper(chan, "h323-ivr-in"));
ast_cdr_setuserfield(chan, cdr_userfield);
if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "cdr userfield set to '%s' on channel '%s'\n",chan->cdr->userfield,chan->name);
chan->priority = 0;
strncpy(chan->exten, numsubst, sizeof(chan->exten)-1);
strncpy(chan->context, new_context, sizeof(chan->context)-1);
if (option_verbose > 2)
	ast_verbose( VERBOSE_PREFIX_3 "Transferring to (%s,%s,%d)\n", chan->context,chan->exten, chan->priority+1);
return 0;

		// create the new channel
		tmp->chan = ast_request("Local", chan->nativeformats, tmpchan);
		if (!tmp->chan) {
			ast_log(LOG_ERROR, "Unable to create local channel for call routing to '%s'\n", tmpchan);
			free(tmp);
			cur = rest;
			continue;
		}
		
ast_log(LOG_ERROR, "Created local channel for call routing to '%s'\n", tmpchan);
		tmp->chan->appl = "AppRDial";
		tmp->chan->data = "(Outgoing Line)";
		tmp->chan->whentohangup = 0;
		if (tmp->chan->callerid)
			free(tmp->chan->callerid);
		if (tmp->chan->ani)
			free(tmp->chan->ani);
		if (chan->callerid)
			tmp->chan->callerid = strdup(chan->callerid);
		else
			tmp->chan->callerid = NULL;
		/* Copy language from incoming to outgoing */
		strcpy(tmp->chan->language, chan->language);
		if (!strlen(tmp->chan->musicclass))
			strncpy(tmp->chan->musicclass, chan->musicclass, sizeof(tmp->chan->musicclass) - 1);
		if (chan->ani)
			tmp->chan->ani = strdup(chan->ani);
		else
			tmp->chan->ani = NULL;
		/* Pass hidecallerid setting */
		tmp->chan->restrictcid = chan->restrictcid;
		/* Pass callingpres setting */
		tmp->chan->callingpres = chan->callingpres;
		/* Presense of ADSI CPE on outgoing channel follows ours */
		tmp->chan->adsicpe = chan->adsicpe;
		/* pass the digital flag */
		ast_dup_flag(tmp->chan, chan, AST_FLAG_DIGITAL);
		/* Place the call, but don't wait on the answer */
ast_log(LOG_ERROR, "Placing call on local channel for call routing to '%s'\n", tmpchan);
		res = ast_call(tmp->chan, numsubst, 0);

		/* Save the info in cdr's that we called them */
		if (chan->cdr) {
			ast_cdr_setdestchan(chan->cdr, tmp->chan->name);
			// Save the returned info from the RADIUS Authentication for use
			// by the RADIUS CDR backend, along with an application tag, the original
			// channels UNIQUEID, and the name of the radius server used so that once we 
			// get to doing the RADIUS accounting, we can determine if the channel for which
			// we are handling the request actually created the call or is a helper channel 
			// (ROUTER or LOCAL channel) and which server needs to handle the accounting
			snprintf(cdr_userfield, sizeof(cdr_userfield),"AppRDial|%s|%s|%s", chan->uniqueid,use_server->Name,pbx_builtin_getvar_helper(chan, "h323-ivr-in"));
			ast_cdr_setuserfield(chan, cdr_userfield);
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "cdr userfield set to '%s' on channel '%s'\n",chan->cdr->userfield,chan->name);
		}
		/* check the restuls of ast_call */
		if (res) {
			/* Again, keep going even if there's an error */
			if (option_debug)
				ast_log(LOG_DEBUG, "ast call on peer returned %d\n", res);
			else if (option_verbose > 2)
				ast_verbose(VERBOSE_PREFIX_3 "Couldn't call %s\n", numsubst);
			ast_hangup(tmp->chan);
			free(tmp);
			cur = rest;
			continue;
		} else
			if (option_verbose > 2)
				ast_verbose(VERBOSE_PREFIX_3 "Called %s\n", numsubst);

		/* Put them in the list of outgoing thingies...  We're ready now. 
		   XXX If we're forcibly removed, these outgoing calls won't get
		   hung up XXX */
		tmp->stillgoing = -1;
		tmp->next = outgoing;
		outgoing = tmp;
		/* If this line is up, don't try anybody else */
		if (outgoing->chan->_state == AST_STATE_UP)
			break;
		cur = rest;
	} while(cur);
	
	if (timeout && strlen(timeout))
		to = atoi(timeout) * 1000;
	else
		to = -1;
ast_log(LOG_ERROR, "Waiting for answer on local channel for call routing\n");
	peer = wait_for_answer(chan, outgoing, &to, &allowredir_in, &allowredir_out, &allowdisconnect);
	if (!peer) {
ast_log(LOG_ERROR, "Outgoing peer did not answer on local channel for call routing\n");
		if (to) 
			/* Musta gotten hung up */
			res = -1;
		 else 
		 	/* Nobody answered, next please? */
			res=0;
		
		goto out;
	}
	if (peer) {
		/* Ah ha!  Someone answered within the desired timeframe.  Of course after this
		   we will always return with -1 so that it is hung up properly after the 
		   conversation.  */
ast_log(LOG_ERROR, "Peer '%s' answered\n", peer->name);
		hanguptree(outgoing, peer);
		outgoing = NULL;
		/* If appropriate, log that we have a destination channel */
		if (chan->cdr) {
			ast_cdr_setdestchan(chan->cdr, peer->name);
			// Save the returned info from the RADIUS Authentication for use
			// by the RADIUS CDR backend
			snprintf(cdr_userfield, sizeof(cdr_userfield),"AppRDial|%s|%s|%s", chan->uniqueid,use_server->Name,pbx_builtin_getvar_helper(chan, "h323-ivr-in"));
			ast_cdr_setuserfield(chan, cdr_userfield);
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "cdr userfield set to '%s' on channel '%s'(2nd time)\n",chan->cdr->userfield,chan->name);
		}
		if (peer->name)
			pbx_builtin_setvar_helper(chan, "DIALEDPEERNAME", peer->name);
		if (numsubst)
			pbx_builtin_setvar_helper(chan, "DIALEDPEERNUMBER", numsubst);
		/* Make sure channels are compatible */
		res = ast_channel_make_compatible(chan, peer);
		if (res < 0) {
			ast_log(LOG_WARNING, "Had to drop call because I couldn't make %s compatible with %s\n", chan->name, peer->name);
			ast_hangup(peer);
			res = -1;
			goto out;
		}

 		/* JDG: sendurl */
 		if( url && strlen(url) && ast_channel_supports_html(peer) ) {
 			ast_log(LOG_DEBUG, "app_rdial: sendurl=%s.\n", url);
 			ast_channel_sendurl( peer, url );
 		} /* /JDG */

		if (announce && announcemsg)
		{
			int res2;
			// Start autoservice on the other chan
			res2 = ast_autoservice_start(chan);
			// Now Stream the File
			if (!res2)
				res2 = ast_streamfile(peer,announcemsg,peer->language);
			if (!res2)
				res2 = ast_waitstream(peer,"");
			// Ok, done. stop autoservice
			res2 = ast_autoservice_stop(chan);
		}
		if (PrePaid == 1) {
			time_t	myt;
			int offset;
			offset = atoi((char *) TimeLimit);
			time(&myt);
        		if (offset)
				chan->whentohangup = myt + offset;
			else
				chan->whentohangup = 0;
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "AbsoluteTimeout set to %d for this call\n",offset);
		}
		res = ast_bridge_call(chan, peer, allowredir_in, allowredir_out, allowdisconnect);

		if (res != AST_PBX_NO_HANGUP_PEER)
			ast_hangup(peer);
	}	
out:
	if (radius_allocated) {
		if (req->vps) {
			pairfree(&req->vps);
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Radius value pairs freed\n");
		}
		if (req->sockfd) {
			close(req->sockfd);
			if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Radius Socket closed\n");
		}
		rad_free(&req);
		req = NULL;
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Radius Packet deallocated\n");
		radius_allocated = 0;
	}

	// free the RADIUS attribute dictionary
//	dict_free();
	
	hanguptree(outgoing, NULL);
	LOCAL_USER_REMOVE(u);
	
	if((go_on>0) && (!chan->_softhangup))
	    res=0;
	    
	return res;
}

static int radius_log(struct ast_cdr *cdr)
{
	RADIUS_PACKET *req;
	RADIUS_PACKET *rep = NULL;
	VALUE_PAIR *list_stop;
	struct radius_server *use_server;
	int id;
	int res;
	struct tm tm;
	time_t t;

	char billsec[32];
	char time_start[80], time_connect[80], time_end[80];
	char radius_info[1024], app_token[128], app_uniqueid[32];
	char call_status[1024];
	char radiusQuery_stop[1024]="";
	char route_provider[10], area_idx[10], route_wholesale[10], route_cost[10];
	char rate_plan[10], billing_unit[10], billing_increment[10];
	char *ptr, *ptr1, *p;
	char tmp[80];

if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "//----- Radius log - channel %s  - UNIQUEID %s ------\n", cdr->channel, cdr->uniqueid);

	ptr = strncpy(radius_info, cdr->userfield, sizeof(radius_info) - 1);
	// parse out the RADIUS application identifier
	ptr1 = strchr(ptr, '|');
	if (ptr1)
		*ptr1 = '\0';
	strncpy(app_token, ptr, sizeof(app_token) - 1);
	if (strcasecmp(app_token, "AppRDial")) {
		ast_log(LOG_ERROR,"unable to do radius accounting - not an RDial initiated call - UNIQUEID: '%s'\n",cdr->uniqueid);
		return -1;
	}
	ptr = ptr1 + 1;
	// parse out the originating calls UNIQUEID
	ptr1 = strchr(ptr, '|');
	if (ptr1)
		*ptr1 = '\0';
	strncpy(app_uniqueid, ptr, sizeof(app_uniqueid) - 1);
	if (strcasecmp(app_uniqueid, cdr->uniqueid)) {
		ast_log(LOG_ERROR,"unable to do radius accounting - This channel is not the originating channel - UNIQUEID: '%s'\n",cdr->uniqueid);
		return -1;
	}
	ptr = ptr1 + 1;
	// parse out the originating calls UNIQUEID
	ptr1 = strchr(ptr, '|');
	if (ptr1)
		*ptr1 = '\0';
	strncpy(tmp, ptr, sizeof(tmp) - 1);
	if (strlen(tmp) == 0) {
		ast_log(LOG_ERROR,"RADIUS server name not specified - unable to do accounting for this call\n");
		return -1;
	}
	use_server = find_server(tmp);

if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Parsing CDR->userfield - %s\n",cdr->userfield);
if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Using RADIUS profile [%s]\n",use_server->Name);
	ptr = ptr1 + 1;
	//parse route_provider
	ptr1 = strchr(ptr, ':');
	if (ptr1)
		*ptr1 = '\0';
	strncpy(route_provider, ptr, sizeof(route_provider) - 1);
	ptr = ptr1 + 1;
	//parse area_idx
	ptr1 = strchr(ptr, ':');
	if (ptr1) 
		*ptr1 = '\0';
	strncpy(area_idx, ptr, sizeof(area_idx) - 1);
	ptr = ptr1 + 1;
	//parse route_wholesale
	ptr1 = strchr(ptr, '=');
	if (ptr1) 
		*ptr1 = '\0';
	strncpy(route_wholesale, ptr, sizeof(route_wholesale) - 1);
	ptr = ptr1 + 1;
	//parse route_cost
	ptr1 = strchr(ptr, ':');
	if (ptr1) 
		*ptr1 = '\0';
	strncpy(route_cost, ptr, sizeof(route_cost) - 1);
	ptr = ptr1 + 1;
	//parse rate_plan
	ptr1 = strchr(ptr, ':');
	if (ptr1) 
		*ptr1 = '\0';
	strncpy(rate_plan, ptr, sizeof(rate_plan) - 1);
	ptr = ptr1 + 1;
	//parse billing_unit & billing_increment
	ptr1 = strchr(ptr, '=');
	if (ptr1)
		*ptr1 = '\0';
	strncpy(billing_unit, ptr, sizeof(billing_unit) - 1);
	strncpy(billing_increment, ptr1+1, sizeof(billing_increment) - 1);
	if ( strlen(billing_unit) == 0 ) {
		strncpy(billing_unit, "30", sizeof(billing_unit) - 1);
	}
	if ( strlen(billing_increment) == 0 ) {
		strncpy(billing_increment, "6", sizeof(billing_increment) - 1);
	}
if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "Parsed RADIUS AUTH INFO for call\n  Route_Provider:\t'%s'\n  area_idx:\t\t'%s'\n  route_wholesale:\t'%s'\n  route_cost:\t\t'%s'\n  rate_plan:\t\t'%s'\n  billing_unit:\t\t'%s'\n  billing_increment:\t'%s'\n",route_provider,area_idx,route_wholesale,route_cost,rate_plan,billing_unit,billing_increment);

//----------------------------------------------------------------------------------------------------------------------------------------

	/* build the radius accounting request and send it to the radius server */
	// because Asterisk only sends CDRs to the CDR backend at the completion of
	// a call, and RADIUS servers require a START and a STOP record, we need to
	// send both in sequence.
	if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "Starting RADIUS START Accounting...\n");
	
	
	if (dict_init(use_server->Dictionary, RADIUS_DICTIONARY) < 0) {
		ast_log(LOG_ERROR, "Unable to initialize RADIUS attribute dictionary\n");
		return -1;
	}

	// set the call status for the call (to place in h323-return-code
	snprintf(call_status,sizeof(call_status),STATUS_FORMAT,"h323-return-code=VoIP",ast_cdr_disp2str(cdr->disposition),route_provider,"1502",area_idx,route_wholesale,route_cost,rate_plan,cdr->dst,billing_unit,billing_increment);

	// convert the start connect and end times to strings
	if (!strcasecmp(use_server->Type, "Cisco")) {
		t = cdr->start.tv_sec;
		localtime_r(&t,&tm);
		if ((cdr->start.tv_usec/1000) < 10) {
			strftime(tmp, sizeof(tmp), DATE_FORMAT_CISCO_1, &tm);
		} else if ((cdr->start.tv_usec/1000) < 100) {
			strftime(tmp, sizeof(tmp), DATE_FORMAT_CISCO_2, &tm);
		} else {
			strftime(tmp, sizeof(tmp), DATE_FORMAT_CISCO_3, &tm);
		}
		snprintf(time_start, sizeof(time_start),tmp,cdr->start.tv_usec/1000);
		t = cdr->answer.tv_sec;
		localtime_r(&t,&tm);
		if ((cdr->answer.tv_usec/1000) < 10) {
			strftime(tmp, sizeof(tmp), DATE_FORMAT_CISCO_1, &tm);
		} else if ((cdr->answer.tv_usec/1000) < 100) {
			strftime(tmp, sizeof(tmp), DATE_FORMAT_CISCO_2, &tm);
		} else {
			strftime(tmp, sizeof(tmp), DATE_FORMAT_CISCO_3, &tm);
		}
		snprintf(time_connect, sizeof(time_connect),tmp,cdr->answer.tv_usec/1000);
		t = cdr->end.tv_sec;
		localtime_r(&t,&tm);
		if ((cdr->end.tv_usec/1000) < 10) {
			strftime(tmp, sizeof(tmp), DATE_FORMAT_CISCO_1, &tm);
		} else if ((cdr->end.tv_usec/1000) < 100) {
			strftime(tmp, sizeof(tmp), DATE_FORMAT_CISCO_2, &tm);
		} else {
			strftime(tmp, sizeof(tmp), DATE_FORMAT_CISCO_3, &tm);
		}
		snprintf(time_end, sizeof(time_end),tmp,cdr->end.tv_usec/1000);
		if ( cdr->answer.tv_sec == 0 && cdr->answer.tv_sec == 0 ) {
			strncpy(time_connect,time_end,sizeof(time_connect));
		}
	} else {
		t = cdr->start.tv_sec;
		localtime_r(&t,&tm);
		strftime(time_start, sizeof(time_start), DATE_FORMAT, &tm);
		t = cdr->answer.tv_sec;
		localtime_r(&t,&tm);
		strftime(time_connect, sizeof(time_connect), DATE_FORMAT, &tm);
		t = cdr->end.tv_sec;
		localtime_r(&t,&tm);
		strftime(time_end, sizeof(time_end), DATE_FORMAT, &tm);
	}
//----------------------------------------------------------------------------------------------------------------------------------------

//do the STOP query
	if (option_verbose > 1) ast_verbose( VERBOSE_PREFIX_3 "Starting RADIUS STOP Accounting...\n");
	if ((req = rad_alloc(1)) == NULL) {
		ast_log(LOG_ERROR, "Unable to allocate memory for RADIUS Accounting STOP request\n");
		return -1;
	} else {
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  RADIUS Accounting STOP Packet allocated...\n");
	}
	id = ((int)getpid() & 0xff);
	req->id = id;
	req->id++;
	if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  RADIUS Accounting STOP request id set to '%d'\n",req->id);
	req->dst_port = use_server->AcctPort;
	req->code = PW_ACCOUNTING_REQUEST;
	req->dst_ipaddr = ip_getaddr(use_server->Server);
	if (req->dst_ipaddr == INADDR_NONE) {
		ast_log(LOG_ERROR, "Unable to resolve address for RADIUS Accounting server '%s'\n", use_server->Server);
		return -1;
	} else {
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  RADIUS Accounting server resolved\n");
	}

	// build the Radius Stop Accounting Query
	snprintf(billsec, sizeof(billsec), "%d", cdr->billsec);
	if (strlen(cdr->accountcode)) {
		snprintf(radiusQuery_stop, 
			sizeof(radiusQuery_stop), 
			"Acct-Session-Id=\"%s\", 
			Acct-Status-Type=\"%s\", 
			h323-setup-time=\"%s\", 
			h323-connect-time=\"%s\", 
			h323-disconnect-time=\"%s\", 
			Acct-Session-Time=\"%s\", 
			user-name=\"%s\", 
			calling-station-id=\"%s\", 
			called-station-id=\"%s\", 
			h323-return-code=\"%s\", 
			Acct-Authentic=\"RADIUS\"\n", 
			cdr->uniqueid, 
			"Stop", 
			time_start, 
			time_connect, 
			time_end, 
			billsec, 
			cdr->accountcode, 
			cdr->src, 
			cdr->dst, 
			call_status);
	} else {
		snprintf(radiusQuery_stop, 
			sizeof(radiusQuery_stop), 
			"Acct-Session-Id=\"%s\", 
			Acct-Status-Type=\"%s\", 
			h323-setup-time=\"%s\", 
			h323-connect-time=\"%s\", 
			h323-disconnect-time=\"%s\", 
			Acct-Session-Time=\"%s\", 
			user-name=\"%s\", 
			calling-station-id=\"%s\", 
			called-station-id=\"%s\", 
			h323-return-code=\"%s\", 
			Acct-Authentic=\"RADIUS\"\n", 
			cdr->uniqueid, 
			"Stop", 
			time_start, 
			time_connect, 
			time_end, 
			billsec, 
			cdr->src, 
			cdr->src, 
			cdr->dst, 
			call_status);
	}

	if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "\nRADIUS Stop QUERY:\n%s", radiusQuery_stop);
	list_stop = NULL;
	p = radiusQuery_stop;
	userparse(p, &list_stop);
	req->vps = list_stop;
if (option_verbose > 2) {
	ast_verbose( VERBOSE_PREFIX_3 " radius value pairs are now:\n");
	vp_printlist(stdout, req->vps);
}
	if ((req->sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		ast_log(LOG_ERROR, "Unable to allocate socket for RADIUS Accounting STOP request\n");
		return -1;
	} else {
		if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  socket allocated for RADIUS Accounting STOP request\n");
	}
	librad_md5_calc(req->vector, req->vector, sizeof(req->vector));
	if (option_verbose > 2) ast_verbose( VERBOSE_PREFIX_3 "  request vector set for RADIUS STOP request\n");
	res = send_radius_acct_request(use_server, req, &rep);
	if (!res) {
		// radius authentication failed
		ast_log(LOG_ERROR,"Unable to perform RADIUS Accounting STOP for call!\n");
		return -1;
	}
	// free the value pair structures
	pairfree(&req->vps);
	//free the request structure
	rad_free(&req);
	//free the RADIUS attribute dictionary
//	dict_free();
	return res;
}

static int load_config(void)
{
	struct ast_config *cfg;
	struct ast_variable *v;
	struct radius_server *s;
	char *cat;
	cfg = ast_load(config);
	if (!cfg) {
		ast_log(LOG_ERROR, "Unable to load config %s, RDial disabled\n", config);
		return 1;
	}

	// parse the global parameters
	v = ast_variable_browse(cfg, "general");
	while(v) {
		if (!strcasecmp(v->name, "Timeout")) {
			RadiusTimeout = atof(v->value);
			if (RadiusTimeout == 0)
			   RadiusTimeout = 1;
		} else if (!strcasecmp(v->name, "Retries")) {
			RadiusRetries = atoi(v->value);
			if (RadiusRetries < 1)
			   RadiusRetries = 1;
		}
		v = v->next;
	}

	// parse the authorization parameters
	cat = ast_category_browse(cfg, NULL);
	while(cat) {
		if (strcasecmp(cat, "general")) {
			s = build_server(cat, ast_variable_browse(cfg, cat));
			if (s) {
				ast_mutex_lock(&radius_list.lock);
				s->next = radius_list.servers;
				radius_list.servers = s;
				ast_mutex_unlock(&radius_list.lock);
				ast_verbose( VERBOSE_PREFIX_3 "Added server profile '%s'\n",s->Name);
			}
		}
		cat = ast_category_browse(cfg, cat);
	}
	ast_destroy(cfg);
	return 0;
}

static int Radius_show_config(int fd, int argc, char *argv[])
{
	int found = 0;
	struct radius_server *s;
	ast_mutex_lock(&radius_list.lock);
	ast_cli(fd, "%s\n\n", "Radius SusSystem Configuration:");
	ast_cli(fd, "%s\n", "General:");
	ast_cli(fd, "\t  %s: %2.2f\n", "Timeout",RadiusTimeout);
	ast_cli(fd, "\t  %s: %i\n\n", "Retries",RadiusRetries);
	ast_cli(fd, "\%s:\n", "PROFILES");
	s = radius_list.servers;
	while(s) {
		found++;
		ast_cli(fd, "[%s]\n", s->Name);
		ast_cli(fd, "\t  %s: %s\n", "Radius Server",s->Server);
		ast_cli(fd, "\t  %s: %s\n", "Radius Server Type",s->Type);
		ast_cli(fd, "\t  %s: %s\n", "Radius Server Secret",s->Secret);
		ast_cli(fd, "\t  %s: %s\n", "Radius Dictionary Directory",s->Dictionary);
		ast_cli(fd, "\t  %s: %d\n", "Radius Authentication Port",s->AuthPort);
		ast_cli(fd, "\t  %s: %d\n", "Radius Accounting Port",s->AcctPort);
		ast_cli(fd, "\t  %s: %2.2f\n", "Timeout",s->Timeout);
		ast_cli(fd, "\t  %s: %d\n\n", "Retries",s->Retries);
		s = s->next;
	}
	ast_mutex_unlock(&radius_list.lock);
	if (found == 0) {
		ast_cli(fd, "NO PROFILES DEFINED");
	}
	ast_cli(fd, "\n");
	return RESULT_SUCCESS;
}

static char show_config_usage[] = 
"Usage: Radius show config\n"
"       Shows the currently configured RADIUS parameters used by the RDial application and cdr_radius backend.\n";

static struct ast_cli_entry  cli_show_config = 
	{ { "radius", "show", "config", NULL }, Radius_show_config, "Show defined Radius configuration", show_config_usage };

int load_module(void)
{
	int res;
	res = load_config();
	if (!res) {
		ast_cli_register(&cli_show_config);
	}
	res = ast_cdr_register(CDR_name, CDR_desc, radius_log);
	if (res) {
		ast_log(LOG_ERROR, "Unable to register CSV CDR handling\n");
	}
	res = ast_register_application(dial_app, RDial_exec, dial_synopsis, dial_descrip);
	res = ast_register_application(set_server_app, set_radius_server, set_server_synopsis, set_server_descrip);
	return res;
}

int unload_module(void)
{
	int res;
	STANDARD_HANGUP_LOCALUSERS;
	ast_log(LOG_NOTICE,"unloading RADIUS CDR backend\n");
	ast_cdr_unregister(CDR_name);
	ast_log(LOG_NOTICE,"unloading RADIUS RDial application\n");
	res = ast_unregister_application(set_server_app);
	if (res == 0) {
		res = ast_unregister_application(dial_app);
	}
	return res;
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
